@@ -42,19 +42,23 @@ __author__ = "Michael Imelfort"
 __copyright__ = "Copyright 2012"
 __credits__ = ["Michael Imelfort"]
 __license__ = "GPL3"
-__version__ = "0.1.6"
+__version__ = "0.2.0"
 __maintainer__ = "Michael Imelfort"
 __email__ = "mike@mikeimelfort.com"
 __status__ = "Development"
 
 ###############################################################################
-import os
 import sys
-import re
+import gzip
 import numpy as np
+import errno
 
 import pysam
-
+from string import maketrans as s_maketrans
+from os.path import splitext as osp_splitext, basename as osp_basename, dirname as osp_dirname, join as osp_join
+from os import makedirs as os_makedirs
+from re import sub as re_sub
+ 
 ###############################################################################
 ###############################################################################
 ###############################################################################
@@ -85,6 +89,7 @@ class BamParser:
                            self.LT.EE:self.LT.EE,
                            self.LT.ERROR:self.LT.ERROR}
         
+        self.compl = s_maketrans('ACGT', 'TGCA')
 #------------------------------------------------------------------------------
 # Managing orientation types
     
@@ -193,7 +198,7 @@ class BamParser:
                 # no need to go here if the type is error
                 if bamType[0] != self.OT.ERROR:
                     # strip off any trailing ".1, _1, /1" which may be at the end of the read id
-                    query = re.sub("[_/\.].$", '', alignedRead.qname)
+                    (query, end) = self.splitReadheader(alignedRead)
                     if query in seen_reads:
                         # we have a pair!
                         # check to see they're on the same strand
@@ -419,6 +424,282 @@ class BamParser:
         else:
             return (rname2, rname1, self.LTInverter[LT], gap)
         
+#------------------------------------------------------------------------------
+# Extract reads from BAM files
+
+    def extractReads(self, bamFiles, 
+                     prefix, 
+                     targets, 
+                     combineBams=False, 
+                     pairsOnly=False, 
+                     combine=False, 
+                     shuffle=False, 
+                     largeFiles=False, 
+                     headersOnly = False,
+                     dontTrustSamFlags=True,
+                     folder='', 
+                     verbose=True):
+        """Extract reads from BAM files
+        
+        targets is a hash of type:
+        { reference_name : bid }
+        
+        If bid == -1 then there is only one bin. This option is used by the command line
+        caller so you can extract using a normal list 
+        """
+        # work out which files have been opened before so we don't nuke stuff!
+        opened_files = {}
+
+        if prefix != '':
+            prefix = "_"+prefix
+
+        # get a collection of storage points
+        read_storage = {}
+        bam_count = 0
+        num_bams = len(bamFiles)
+        for bf in bamFiles:
+            seen_reads = {}
+            read_storage[bam_count] = {}
+            for bid in targets.values():
+                read_storage[bam_count][bid] = [[],[],[]] # paired1, paired2, singles
+            try:
+                bam_file = pysam.Samfile(bf, 'rb')
+                if verbose:
+                    print "    Parsing BAM '%s' (%d of %d)" % (getBamDescriptor(bf), (bam_count+1), num_bams)
+                
+                # use these to work out if we need quality scores or not
+                w_qual = 0
+                wo_qual = 0
+                
+                # now get the reads associated with each target
+                for reference, length in zip(bam_file.references, bam_file.lengths):
+                    rl = ReadLoader()
+                    bam_file.fetch(reference, 0, length, callback = rl )
+                    for alignedRead in rl.alignedReads:
+                        (query, end) = self.splitReadheader(alignedRead, dontTrustSamFlags=dontTrustSamFlags)
+                        # we have basically thrown out all pairing information here
+                        # so we make some assumptions:
+                        #
+                        # The first read we see in a pair is the first read
+                        # We trust the reversed flag
+                        # We don't mess with the relative orientation that follows
+                        #
+                        if query in seen_reads:
+                            # get the raw reads
+                            if headersOnly:
+                                # no need to store all the information as we'll just be nuking it anyway
+                                read1 = alignedRead.qname
+                                read2 = seen_reads[query][0].qname
+                            else:
+                                if seen_reads[query][0].is_reverse:
+                                    seen_read = (seen_reads[query][0].qname, self.revComp(seen_reads[query][0].seq), seen_reads[query][0].qual)
+                                else:
+                                    seen_read = (seen_reads[query][0].qname, seen_reads[query][0].seq, seen_reads[query][0].qual)
+                                if alignedRead.is_reverse:
+                                    align_read = (alignedRead.qname, self.revComp(alignedRead.seq), alignedRead.qual)
+                                else:
+                                    align_read = (alignedRead.qname, alignedRead.seq, alignedRead.qual)
+    
+                                # work out which is the first in the pair and which is the second
+                                if end == 1:
+                                    read1 = align_read
+                                    read2 = seen_read
+                                else:
+                                    read1 = seen_read
+                                    read2 = align_read
+
+                                # check for quality info
+                                if read1[2] is None:
+                                    wo_qual += 1
+                                else:
+                                    w_qual += 1
+                            
+                            # put them in the storage
+                            try:
+                                bid1 = targets[bam_file.getrname(alignedRead.tid)]
+                                read_storage[bam_count][bid1][0].append(read1)
+                                read_storage[bam_count][bid1][1].append(read2)
+                            except KeyError:
+                                pass
+                            try:
+                                bid2 = targets[bam_file.getrname(seen_reads[query][0].tid)]
+                                if bid2 != bid1:
+                                    read_storage[bam_count][bid2][0].append(read1)
+                                    read_storage[bam_count][bid2][1].append(read2)
+                            except KeyError:
+                                pass
+                            
+                            # delete this guy so we can determine who the pairs are at the end
+                            del seen_reads[query]
+                        else:
+                            seen_reads[query] = (alignedRead, end)
+                
+                for ar in seen_reads:
+                    alignedRead = seen_reads[ar][0]
+                    if headersOnly:
+                        read = alignedRead.qname
+                    else:
+                        if alignedRead.is_reverse:
+                            read = (alignedRead.qname, self.revComp(alignedRead.seq), alignedRead.qual)
+                        else:
+                            read = (alignedRead.qname, alignedRead.seq, alignedRead.qual)
+                    try:
+                        bid = targets[bam_file.getrname(alignedRead.tid)]
+                        read_storage[bam_count][bid][2].append(read)
+                    except KeyError:
+                        pass
+
+                bam_file.close()
+                
+                # now we can write to file
+                # work out file extension
+                has_qual = w_qual > wo_qual
+                if headersOnly:
+                    extension = 'headers'
+                    shuffle = True # force everything into one file
+                    pairsOnly = True
+                else:
+                    if has_qual:
+                        extension = "fq"
+                    else:
+                        extension = "fa"
+                
+                # work out compression
+                if largeFiles:
+                    fopen = open
+                else:
+                    fopen = gzip.open
+                    extension += '.gz'
+
+                # get a basename
+                if combineBams:
+                    # always use the same file
+                    base_name = "extracted"
+                else:
+                    # need a different base name for each bam
+                    base_name = getBamDescriptor(bf)
+
+                if folder == '':
+                    # use the bam file's folder
+                    base_name = getBamStem(bf)
+                else:
+                    makeSurePathExists(folder)
+                    base_name = osp_join(folder, base_name)
+
+                unique_bids = {}
+                for bid in targets.values():
+                    try:
+                        unique_bids[bid] += 1
+                    except KeyError:
+                         unique_bids[bid] = 1
+
+                pretty_counter = 0    
+                print "    ",                    
+                for bid in unique_bids:
+                    print ".",
+                    sys.stdout.flush()
+                    pretty_counter += 1
+                    if pretty_counter > 79:
+                        pretty_counter = 0
+                        print "\n    ",
+                    if bid == -1:
+                        # single bin, no need to be fancy
+                        if shuffle:
+                            # if we are going to shuffle reads then we don't need
+                            # to open a second file handle
+                            read1_fh = self.openFileCorrectly("%s%s.%s" % (base_name, prefix, extension), fopen, opened_files)
+                            read2_fh = read1_fh
+                        else:
+                            read1_fh = self.openFileCorrectly("%s%s_1.%s" % (base_name, prefix, extension), fopen, opened_files)
+                            read2_fh = self.openFileCorrectly("%s%s_2.%s" % (base_name, prefix, extension), fopen, opened_files)
+                    else:
+                        if shuffle:
+                            read1_fh = self.openFileCorrectly("%s%s_BIN_%d.%s" % (base_name, prefix, bid, extension), fopen, opened_files)
+                            read2_fh = read1_fh
+                        else:
+                            read1_fh = self.openFileCorrectly("%s%s_BIN_%d_1.%s" % (base_name, prefix, bid, extension), fopen, opened_files)
+                            read2_fh = self.openFileCorrectly("%s%s_BIN_%d_2.%s" % (base_name, prefix, bid, extension), fopen, opened_files)
+                            
+                    close_unpaired_file = False
+                    if not pairsOnly:
+                        # we need to write the unpaired guys somewhere
+                        if combine and shuffle:
+                            # we can just add them to the end of read1_fh
+                            unpaired_fh = read1_fh
+                        else:
+                            # either the user wanted things separate or they chose
+                            # not to shuffle, either way a new fh is needed
+                            close_unpaired_file = True
+                            if bid == -1:
+                                unpaired_fh = self.openFileCorrectly("%s%s_unpaired.%s" % (base_name, prefix, extension), fopen, opened_files)
+                            else:
+                                unpaired_fh = self.openFileCorrectly("%s%s_BIN_%d_unpaired.%s" % (base_name, prefix, bid, extension), fopen, opened_files)
+                    
+                    # now we print
+                    if headersOnly:
+                        headers = {} # pipe into a hash first so we don't get dupes
+                        for i in range(len(read_storage[bam_count][bid][0])):
+                            headers[read_storage[bam_count][bid][0][i]] = True
+                            headers[read_storage[bam_count][bid][1][i]] = True
+                        for i in range(len(read_storage[bam_count][bid][2])):
+                            headers[read_storage[bam_count][bid][2][i]] = True
+                        for header in headers:
+                            read1_fh.write(header+"\n")
+                    else:
+                        for i in range(len(read_storage[bam_count][bid][0])):
+                            read1_fh.write(self.formatRead(read_storage[bam_count][bid][0][i], has_qual))
+                            read2_fh.write(self.formatRead(read_storage[bam_count][bid][1][i], has_qual))
+                        if not pairsOnly:
+                            for i in range(len(read_storage[bam_count][bid][2])):
+                                unpaired_fh.write(self.formatRead(read_storage[bam_count][bid][2][i], has_qual))
+                
+                    read1_fh.close()
+                    if not shuffle:
+                        read2_fh.close()
+                    if close_unpaired_file:
+                        unpaired_fh.close()
+                print "\n",
+                
+            except:
+                print "Unable to open BAM file",bf,"-- did you supply a SAM file instead?"
+                raise
+            bam_count += 1
+
+    def openFileCorrectly(self, fileName, fopen, openedFiles):
+        """open a file in write mode or append mode depending on it if's been opened before"""
+        try:
+            open_mode = openedFiles[fileName]
+        except KeyError:
+            open_mode = 'w'
+            openedFiles[fileName] = 'a'
+        return fopen(fileName, open_mode)
+        
+            
+    def formatRead(self, readInfo, fastq):
+        """Make a formatted string ready for printing"""
+        if fastq:
+            return "@%s\n%s\n+\n%s\n" % (readInfo[0], readInfo[1], readInfo[2])
+        else:
+            return ">%s\n%s\n" % (readInfo[0], readInfo[1])
+        
+    def splitReadheader(self, alignedRead, dontTrustSamFlags=False):
+        """Split the read header to determine the common part of a pair and it's pair ref"""
+        query = re_sub("[_/\.][12]$", '', alignedRead.qname) # this will not affect illumina headers
+        if dontTrustSamFlags:
+            # try work out pairing info from the read header itself
+            try:
+                end = int(re_sub(".*[_/\.]",'', alignedRead.qname))
+            except:
+                return (query, None)
+            if end != 1 and end != 2:
+                return (query, None)
+        else:
+            if alignedRead.is_read1:
+                end = 1
+            else:
+                end = 2
+        return (query, end)
+
 #------------------------------------------------------------------------------
 # Working out read types
 
@@ -464,7 +745,7 @@ class BamParser:
                 if(num_stored > numPaired):
                     break                               # we have enough
                 # strip off any trailing ".1, _1, /1" which may be at the end of the read id
-                query = re.sub("[_/\.].$", '', alignedRead.qname)
+                (query, end) = self.splitReadheader(alignedRead)
                 if query in seen_reads:
                     # we have a pair!
                     # check to see they're on the same strand
@@ -515,15 +796,31 @@ class BamParser:
             else:
                 # --1-> --2->
                 return (self.OT.SAME, isize)
-    
+
+    def revComp(self, seq):
+        """Return the reverse complement of a sequence"""
+        # build a dictionary to know what letter to switch to
+        return seq.translate(self.compl)[::-1]
+
 def getBamDescriptor(fullPath):
     """AUX: Reduce a full path to just the file name minus extension"""
-    return os.path.splitext(os.path.basename(fullPath))[0]
+    return osp_splitext(osp_basename(fullPath))[0]
+
+def getBamStem(fullPath):
+    """AUX: Reduce a full path to the path + file name minus extension"""
+    return osp_splitext(fullPath)[0]
 
 def enum(*sequential, **named):
     enums = dict(zip(sequential, range(len(sequential))), **named)
     return type('Enum', (), enums)
 
+def makeSurePathExists(path):
+    try:
+        os_makedirs(path)
+    except OSError as exception:
+        if exception.errno != errno.EEXIST:
+            raise
+
 ###############################################################################
 ###############################################################################
 ###############################################################################
